#ifdef _WIN32
#include <windows.h>
#endif // _WIN32

//#include <GL/gl.h>
//#include <GL/glu.h>

#ifdef __APPLE__
#include <GLUT/glut.h>
#else
#include <GL/glut.h>
#endif

#include <iostream>
#include <string>

#include <boost/shared_ptr.hpp>

#include <PinholeCamera.h>
#include <ErrorManager.h>
#include <GageAdaptor.h>
#include <IMarchingMethod.h>
#include <MarchingCubes.h>
#include <MarchingTetrahedra.h>
#include <DualContouring.h>
#include <MarchingCubes.h>
#include <VertexPool.h>
#include <Timer.h>
#include <MeshQuality.h>

using namespace std;

// Image.
boost::shared_ptr<CGageAdaptor> m_image;
// Polygonizer.
boost::shared_ptr<IMarchingMethod> m_polygonizer;
float m_isoLevel = 0.0f;
bool m_drawEdges = false;
bool m_drawActiveEdges = false;
// Quality measurer.
boost::shared_ptr<CMeshQuality> m_qualityMeasurer;
// Reference grid.
unsigned int  m_gridDisplayList;
// Camera.
CPinholeCamera m_camera;
int m_mouseButton = 0;
int m_mouseX = -1;
int m_mouseY = -1;
// Polygonizer
IMarchingMethod::MARCHING_METHOD m_method;
IMarchingMethod::EDGE_TRANSFORM m_edgeTransform;

IMarchingMethod *m_marchingCubes=NULL, *m_marchingTetrahedra=NULL, *m_dualContouring=NULL;
//boost::shared_ptr<CMarchingCubes> m_marchingCubes;
//boost::shared_ptr<CMarchingTetrahedra> m_marchingTetrahedra;
//boost::shared_ptr<CDualContouring> m_dualContouring;

/**
*/
bool LoadImage(const string& path)
{
	double kernelParam[3];

	// Scale parameter, in units of samples.
    kernelParam[0] = 1.0; 
	// B.
    kernelParam[1] = 0.0;
	// C.
    kernelParam[2] = 0.5;

	m_image.reset(new CGageAdaptor);

	if (!m_image.get())
	{
		MarkError();

		return false;
	}

	if (!m_image->Open(path))
	{
		MarkError();

		return false;
	}

	// BC family of cubic polynomial splines.
	if (!m_image->SetValueKernel(nrrdKernelBCCubic, kernelParam))
	{
		MarkError();

		return false;
	}

	// 1st deriv. of BC cubic family.
	if (!m_image->Set1stDerivativeKernel(nrrdKernelBCCubicD, kernelParam))
	{
		MarkError();

		return false;
	}

	return true;
}

/**
*/
bool GetImageScale(float *sx, float *sy, float *sz)
{
	float majorAxisSize;

    majorAxisSize = (float)max(m_image->GetWidth(), max(m_image->GetHeight(), m_image->GetDepth()));
	
    *sx = 20.0f/majorAxisSize;
    *sy = 20.0f/majorAxisSize;
    *sz = 20.0f/majorAxisSize;

	return true;
}

/**
(test purposes only).
*/
bool MapQualityToColorArray(boost::shared_ptr<CVertexPool> mesh, boost::shared_ptr<CMeshQuality> qualityMeasurer)
{
	float colorTable[] = {
		1.0f, 0.0f, 1.0f,
		0.0f, 0.0f, 1.0f,
		0.0f, 1.0f, 1.0f,
		0.0f, 1.0f, 0.0f,
		1.0f, 1.0f, 0.0f,
		1.0f, 0.0f, 0.0f
	};
	int firstColor,
		secondColor,
		i,
		j;
	float vertexQuality,
		mapping,
		t;

	mesh->m_colorArray.clear();
	mesh->m_colorSize = 3;

	mesh->m_colorArray.reserve(3*mesh->m_vertexArray.size());

	for (j=0; j!=(int)mesh->m_vertexArray.size()/mesh->m_vertexSize; j++)
	{
		vertexQuality = qualityMeasurer->GetVertexQuality(j);

		mapping = vertexQuality*5.0f;

		firstColor = (int)mapping;

		if (firstColor == 5)
		{
			t = 0.0f;

			firstColor--;
		}
		else
			t = mapping - firstColor;

		secondColor = firstColor + 1;

		for (i=0; i!=3; i++)
			mesh->m_colorArray.push_back(*(colorTable + 3*firstColor + i)*(1.0f - t) + *(colorTable + 3*secondColor + i)*t);
	}

	return true;
}

/**
*/
bool ResetMarchingMethod() {
	IMarchingMethod *marchingMethod;
	switch (m_method) {
	case IMarchingMethod::MARCHING_CUBES:
		if (m_marchingCubes == NULL)
			m_marchingCubes = new CMarchingCubes;
		marchingMethod = m_marchingCubes;
		break;
	case IMarchingMethod::MARCHING_TETRAHEDRA:
		if (m_marchingTetrahedra == NULL)
			m_marchingTetrahedra = new CMarchingTetrahedra;
		marchingMethod = m_marchingTetrahedra;
		break;
	case IMarchingMethod::DUAL_CONTOURING:
		if (m_dualContouring == NULL)
			m_dualContouring = new CDualContouring;
		marchingMethod = m_dualContouring;
		break;
	}
	marchingMethod->SetMethod(m_method);
		
	if (!marchingMethod)
	{
		MarkError();

		return false;
	}
	
	if (!marchingMethod->Load(m_image))
	{
		MarkError();

		return false;
	}

	if (!marchingMethod->SetGridSize(m_image->GetWidth(), m_image->GetHeight(), m_image->GetDepth()))
	{
		MarkError();

		return false;
	}

	m_polygonizer.reset(marchingMethod);

	if (!m_polygonizer.get())
	{
		MarkError();

		return false;
	}
	

	m_qualityMeasurer.reset(new CMeshQuality);

	if (!m_qualityMeasurer.get())
	{
		MarkError();

		return false;
	}

}

/**
*/
bool GenerateIsosurface(float isoLevel)
{
	CTimer timer;

	if (!m_polygonizer.get())
	{
		IMarchingMethod *marchingMethod;
		switch (m_method) {
		case IMarchingMethod::MARCHING_CUBES:
			marchingMethod = new CMarchingCubes; 
			break;
		case IMarchingMethod::MARCHING_TETRAHEDRA: 
			marchingMethod = new CMarchingTetrahedra; 
			break;
		case IMarchingMethod::DUAL_CONTOURING: 
			marchingMethod = new CDualContouring; 
			break;
		}
		marchingMethod->SetMethod(m_method);
		
		if (!marchingMethod)
		{
			MarkError();

			return false;
		}

		if (!marchingMethod->Load(m_image))
		{
			MarkError();

			return false;
		}

		if (!marchingMethod->SetGridSize(m_image->GetWidth(), m_image->GetHeight(), m_image->GetDepth()))
		{
			MarkError();

			return false;
		}

		m_polygonizer.reset(marchingMethod);

		if (!m_polygonizer.get())
		{
			MarkError();

			return false;
		}

		m_qualityMeasurer.reset(new CMeshQuality);

		if (!m_qualityMeasurer.get())
		{
			MarkError();

			return false;
		}
	}

	switch (m_method) {
	case IMarchingMethod::MARCHING_CUBES:
		cout << "Marching_Method = MARCHING CUBES\n" ;
		break;
	case IMarchingMethod::MARCHING_TETRAHEDRA: 
		cout << "Marching_Method = MARCHING TETRAHEDRA\n" ;
		break;
	case IMarchingMethod::DUAL_CONTOURING: 
		cout << "Marching_Method = DUAL CONTOURING\n" ;
		break;
	}

	cout << "Generating isosurface at level " << isoLevel << "... ";

	m_polygonizer->SetTransform(m_edgeTransform);
	switch (m_edgeTransform) 
	{
	case IMarchingMethod::NONE:
		break;
	case IMarchingMethod::GRADIENT_TRANSFORM:
		cout << "(using gradient transform)";
		break;
	case IMarchingMethod::TANGENTIAL_TRANSFORM:
		cout << "(using tangential transform)";
		break;
	}

	timer.Restart();

	if (!m_polygonizer->GenerateMesh(isoLevel))
	{
		MarkError();

		return false;
	}

	cout << "ok (done in " << timer.GetElapsed() << " ms)." << endl;

	m_qualityMeasurer->Load(m_polygonizer->GetMesh());

	m_qualityMeasurer->MeasureQuality();

	MapQualityToColorArray(m_polygonizer->GetMesh(), m_qualityMeasurer);

	cout << "Min/max/avg quality: " << m_qualityMeasurer->GetMinQuality() << "/" << m_qualityMeasurer->GetMaxQuality() << "/" << m_qualityMeasurer->GetAverageQuality() << "." << endl;

	return true;
}

/**
*/
void DrawGrid(void)
{
    if (glIsList(m_gridDisplayList) == GL_TRUE)
        glCallList(m_gridDisplayList);
    else
    {
        m_gridDisplayList = glGenLists(1);

        glNewList(m_gridDisplayList, GL_COMPILE_AND_EXECUTE);

        glLineStipple(1, 0xaaaa);

        glEnable(GL_LINE_STIPPLE);

        glColor3f(0.5f, 0.5f, 0.5f);

        glBegin(GL_LINES);

        for (unsigned int i=0; i!=11; i++)
        {
            // yz
            glVertex3i( 0,  i,  0);
            glVertex3i( 0,  i, 10);

            glVertex3i( 0,  0,  i);
            glVertex3i( 0, 10,  i);

            // xz
            glVertex3i( i,  0,  0);
            glVertex3i( i,  0, 10);

            glVertex3i( 0,  0,  i);
            glVertex3i(10,  0,  i);

            // xy
            glVertex3i( i,  0,  0);
            glVertex3i( i, 10,  0);

            glVertex3i( 0,  i,  0);
            glVertex3i(10,  i,  0);
        }

        glEnd();

        glDisable(GL_LINE_STIPPLE);

        glLineWidth(2.0f);

        glBegin(GL_LINES);

        glColor3f(1.0f, 0.0f, 0.0f);
        glVertex3i( 0,  0,  0);
        glVertex3i(10,  0,  0);

        glColor3f(0.0f, 1.0f, 0.0f);
        glVertex3i( 0,  0,  0);
        glVertex3i( 0, 10,  0);

        glColor3f(0.0f, 0.0f, 1.0f);
        glVertex3i( 0,  0,  0);
        glVertex3i( 0,  0, 10);

        glEnd();

        glLineWidth(1.0f);

        glEndList();
    }
}

/**
*/
void DrawHelp(void)
{
	cout << endl;
	cout << "Available commands:" << endl;

	cout << "ESC:\t" << "exit program." << endl;
	cout << "'a':\t" << "enable/disable active edges drawing." << endl;
	cout << "'e':\t" << "enable/disable mesh edges drawing." << endl;
	cout << "'g':\t" << "enable the gradient transform and (re)generates the polygonization." << endl;
	cout << "'h':\t" << "Draw this help." << endl;
	cout << "'n':\t" << "disable the edge transforms and (re)generates the polygonization." << endl;
	cout << "'t':\t" << "enable the tangential transform and (re)generates the polygonization." << endl;
	cout << endl;
}

/**
*/
void EnableLight(void)
{
	float lightPosition[] =	{0.0f, 0.0f, -100.0f, 0.0f}, 
		lightDiffuse[] = {0.8f, 0.8f, 0.8f, 1.0f};

    glLightModeli(GL_LIGHT_MODEL_TWO_SIDE, GL_TRUE);

	glMatrixMode(GL_MODELVIEW);

	glPushMatrix();
	
	glLoadIdentity();

	glLightfv(GL_LIGHT0, GL_POSITION, lightPosition);

	glPopMatrix();

	glLightfv(GL_LIGHT0, GL_DIFFUSE, lightDiffuse);

	glEnable(GL_LIGHTING);
	glEnable(GL_LIGHT0);
}

/**
*/
void DisableLight(void)
{
	glDisable(GL_LIGHTING);
}

/**
*/
void OnDestroy(void)
{
	int windowName;

    if (glIsList(m_gridDisplayList) == GL_TRUE)
        glDeleteLists(m_gridDisplayList, 1);

	windowName = glutGetWindow();

	if (windowName)
		glutDestroyWindow(windowName);

	if (!CErrorManager::Instance().IsOk())
		CErrorManager::Instance().FlushErrors();

	exit(EXIT_SUCCESS);
}

/**
*/
void OnPaint(void)
{
	float scaleX,
		scaleY,
		scaleZ;

    glClearColor(1.0f, 1.0f, 1.0f, 1.0f);

    glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);

    m_camera.ApplyTransform();

    glEnable(GL_DEPTH_TEST);
	
	//DrawGrid();

	if (m_polygonizer->IsValid())
	{
		glTranslatef(-10.0f, -10.0f, -10.0f);

		GetImageScale(&scaleX, &scaleY, &scaleZ);

		glScalef(scaleX, scaleY, scaleZ);
	    
		EnableLight();

		glEnable(GL_NORMALIZE);

		glEnable(GL_COLOR_MATERIAL);

        glPolygonOffset(1.0, 1.0);

		glEnable(GL_POLYGON_OFFSET_FILL);
		
		glColor3f(1.0f, 1.0f, 0.0f);

		m_polygonizer->GetMesh()->Draw();

        glDisable(GL_POLYGON_OFFSET_FILL);

		glDisable(GL_COLOR_MATERIAL);

		glDisable(GL_NORMALIZE);

		DisableLight();

        if (m_drawEdges)
        {
            glColor3f(0.3f, 0.3f, 0.3f);

            glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);

			glVertexPointer(m_polygonizer->GetMesh()->m_vertexSize, GL_FLOAT, 0, &m_polygonizer->GetMesh()->m_vertexArray[0]);

			glEnableClientState(GL_VERTEX_ARRAY);

			glDrawElements(m_polygonizer->GetMesh()->m_mode, (int)m_polygonizer->GetMesh()->m_elementArray.size(), GL_UNSIGNED_INT, &m_polygonizer->GetMesh()->m_elementArray[0]);

			glDisableClientState(GL_VERTEX_ARRAY);

            glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
        }

		if (m_drawActiveEdges)
        {
            glColor3f(1.0f, 0.0f, 0.0f);

			m_polygonizer->GetActiveEdges()->Draw();
        }
	}

    glDisable(GL_DEPTH_TEST);
	
    glutSwapBuffers();
}

/**
*/
void OnSize(int w, int h)
{
    if (h == 0)
        h = 1;

    glViewport(0, 0, w, h);

    m_camera.Create(45.0f, 1.0f, 50.0f, w, h);
}

/**
*/
void OnIdle(void)
{
    //glutPostRedisplay();
}

/**
*/
void OnVisible(int isVisible)
{
    if (isVisible == GLUT_VISIBLE)	
        glutIdleFunc(OnIdle);
    else 
        glutIdleFunc(0);
}

/**
*/
void OnKeyDown(unsigned char key, int x, int y) 
{
    switch (key)
    {
	case 'a':
	case 'A':
		m_drawActiveEdges = !m_drawActiveEdges;
		break;
	case 'c':
	case 'C':
		if (m_method != IMarchingMethod::MARCHING_CUBES) {
			m_method = IMarchingMethod::MARCHING_CUBES;
			if (!GenerateIsosurface(m_isoLevel))
			{
				MarkError();

				OnDestroy();
			}
			ResetMarchingMethod();
		}
		break;		
	case 'd':
	case 'D':
		if (m_method != IMarchingMethod::DUAL_CONTOURING) {
			m_method = IMarchingMethod::DUAL_CONTOURING;
			if (!GenerateIsosurface(m_isoLevel))
			{
				MarkError();

				OnDestroy();
			}
			ResetMarchingMethod();
		}
		break;		
	case 'e':
	case 'E':
		m_drawEdges = !m_drawEdges;
		break;
	case 'g':
	case 'G':
		m_edgeTransform = IMarchingMethod::GRADIENT_TRANSFORM;
		if (!GenerateIsosurface(m_isoLevel))
		{
			MarkError();

			OnDestroy();
		}
		break;
	case 'h':
	case 'H':
		DrawHelp();
		break;
	case 'n':
	case 'N':
		m_edgeTransform = IMarchingMethod::NONE;
		if (!GenerateIsosurface(m_isoLevel))
		{
			MarkError();

			OnDestroy();
		}
		break;
    case 'q':
    case 'Q':
    case 27:
        OnDestroy();

        exit(EXIT_SUCCESS);
        break;
	case 'r':
	case 'R':
		m_edgeTransform = IMarchingMethod::TANGENTIAL_TRANSFORM;
		if (!GenerateIsosurface(m_isoLevel))
		{
			MarkError();

			OnDestroy();
		}
		break;
	case 't':
	case 'T':
		if (m_method != IMarchingMethod::MARCHING_TETRAHEDRA) {	
			m_method = IMarchingMethod::MARCHING_TETRAHEDRA;
			if (!GenerateIsosurface(m_isoLevel))
			{
				MarkError();

				OnDestroy();
			}
			ResetMarchingMethod();
		}
		break;		
    }

    glutPostRedisplay();
}

/**
*/
void OnMouse(int button, int state, int x, int y)
{
    if (state == GLUT_DOWN)
        m_mouseButton = button;
    else
        m_mouseButton = 0;

    m_mouseX = x;
    m_mouseY = y;
}

/**
*/
void OnMouseMove(int x, int y)
{
    switch (m_mouseButton)
    {
    case GLUT_LEFT_BUTTON:
        m_camera.Pitch(0.1f*(y - m_mouseY));
        m_camera.Yaw(0.1f*(x - m_mouseX));
        break;

    case GLUT_MIDDLE_BUTTON:
        m_camera.MoveSide(0.025*(x - m_mouseX));
        m_camera.MoveUp(0.025f*(m_mouseY - y));
        break;

    case GLUT_RIGHT_BUTTON:
        m_camera.MoveFront(0.05f*(y - m_mouseY));
        break;
    }

    m_mouseX = x;
    m_mouseY = y;

    glutPostRedisplay();
}

/**
*/
void OnCreate(int argc, char** argv)
{
    glutInit(&argc, argv);
    glutInitDisplayMode(GLUT_DOUBLE|GLUT_RGBA|GLUT_DEPTH);
    glutInitWindowSize(512, 512);

    glutCreateWindow("Test CSimpleMarchingCubes implementation.");

    m_camera.Create(45.0f, 1.0f, 50.0f, 512, 512);

    m_camera.MoveFront(-20.0f);
    m_camera.ApplyTransform();  

    glutDisplayFunc(OnPaint);
    glutKeyboardFunc(OnKeyDown);
    glutMouseFunc(OnMouse);
    glutMotionFunc(OnMouseMove);
    glutReshapeFunc(OnSize);
    glutVisibilityFunc(OnVisible);
}

/**
*/
int main(int argc, char **argv)
{
	if (argc != 6)
	{
		cout << endl;
		cout << "SYPNOSIS" << endl;
		cout << "TestMarchingCubes SOURCE.nhdr ISOLEVEL marching-method (c,d,t) transformation (0,g,t) output-file" << endl;
		cout << endl;

		OnDestroy();
	}

	DrawHelp();

// Argument argv[1] has the .nrrd filename
	if (!LoadImage(argv[1]))
	{
		MarkError();

		OnDestroy();
	}

// Argument argv[2] has the isovalue
	m_isoLevel = atof(argv[2]);

// Argument argv[3] defines which method to use for polygonization:
// c - marching cubes, t - marching tetrahedra, d - dual contouring

// Argument argv[4] defines if edge transformations will be applied
// 0 - none, g - gradient transform, t - tangential transform
	switch (argv[3][0]) {
	case 'c': m_method = IMarchingMethod::MARCHING_CUBES; break;
	case 't': m_method = IMarchingMethod::MARCHING_TETRAHEDRA; break;
	case 'd': m_method = IMarchingMethod::DUAL_CONTOURING; break;
	default: 
		cout << "\n invalid selection of algorithm (c (marching cubes) t (marching tetra) d (dc) expected)";
	}
	switch (argv[4][0]) {
	case '0': m_edgeTransform = IMarchingMethod::NONE; break;
	case 'g': m_edgeTransform = IMarchingMethod::GRADIENT_TRANSFORM; break;
	case 't': m_edgeTransform = IMarchingMethod::TANGENTIAL_TRANSFORM; break;
	default: 
		cout << "\n invalid selection of edge transform (0 (none) g (gradient) t (tangential) expected)";
	}

#if 1
	GenerateIsosurface(m_isoLevel);
	if (m_polygonizer->IsValid()) 
		cout << "\nMesh is valid - OFF generated\n";
	else
		cout << "\nMesh is invalid\n";
	m_polygonizer->GetMesh()->Save (argv[5], CVertexPool::PLY, true);
#else
	if (!GenerateIsosurface(m_isoLevel))
	{
		MarkError();

		OnDestroy();
	}

    OnCreate(argc, argv);

    glutMainLoop();
#endif

    return EXIT_SUCCESS;
}

