/* Copyright 2007 Universidade Federal do Rio Grande do Sul
   Carlos Dietrich - cadietrich@inf.ufrgs.br

   This copy is GPLv2 or later, your choice. */

#ifndef GAGEADAPTOR_INCLUDED
#define GAGEADAPTOR_INCLUDED

#include <string>

#include <teem/nrrd.h>
#include <teem/gage.h>

#if TEEM_VERSION_MINOR < 10
#define GAGE_TYPE float
#else
#define GAGE_TYPE double
#endif

#include <boost/utility.hpp>

class CGageAdaptor
	: boost::noncopyable
{
public:
	enum QUERY_ITEM {
		// Data value.
		VALUE = gageSclValue,
		// Gradient vector, normalized.
		NORMAL = gageSclNormal,
		// Gradient vector, un-normalized.
		GRADIENT = gageSclGradVec,
		// Gradient magnitude.
		GRADIENT_MAGNITUDE = gageSclGradMag,
		// Hessian (column-order).
		HESSIAN = gageSclHessian,
		// Laplacian: Dxx + Dyy + Dzz.
		LAPLACIAN = gageSclLaplacian,
		// Hessian's 1st eigenvalue.
		HESSIAN_1ST_EIGENVALUE = gageSclHessEval0,
		// Hessian's 2nd eigenvalue.
		HESSIAN_2ND_EIGENVALUE = gageSclHessEval1,
		// Hessian's 3rd eigenvalue.
		HESSIAN_3RD_EIGENVALUE = gageSclHessEval2,
		// 1st principle curvature.
		PRINCIPLE_CURVATURE = gageSclK1
	};
	CGageAdaptor(void);
	CGageAdaptor(const std::string& path);
	virtual ~CGageAdaptor(void);
	bool Open(const std::string& path);
	void Close(void);
	bool IsOpen(void) const;
	bool EnableQuery(int item);
	bool ResetKernel(void);
	void SetClamp(bool doClamp);
	bool SetValueKernel(const NrrdKernel *type, const double *parameters);
	bool Set1stDerivativeKernel(const NrrdKernel *type, const double *parameters);
	bool Set2ndDerivativeKernel(const NrrdKernel *type, const double *parameters);
	int GetWidth(void) const;
	int GetHeight(void) const;
	int GetDepth(void) const;
	const GAGE_TYPE *GetValueArray(void) const;
	float GetValue(float x, float y, float z) const;
	const GAGE_TYPE *GetNormal(float x, float y, float z) const;
	const GAGE_TYPE *GetGradient(float x, float y, float z) const;
	float GetGradientMagnitude(float x, float y, float z) const;
	const GAGE_TYPE *GetHessian(float x, float y, float z) const;
	float GetLaplacian(float x, float y, float z) const;
	float GetHessian1stEigenvalue(float x, float y, float z) const;
	float GetHessian2ndEigenvalue(float x, float y, float z) const;
	float GetHessian3rdEigenvalue(float x, float y, float z) const;
	float Get1stPrincipleCurvature(float x, float y, float z) const;
private:
	bool OpenNrrd(const std::string& path);
	bool CreateDefaultContext(void);
	bool UpdateKernel(void);
	inline void Clamp(float *x, float *y, float *z) const;
	void Create(void);
protected:
	Nrrd *m_imageHandle;
	gageContext *m_measurementContext;
	gagePerVolume *m_imageInfo;
	const GAGE_TYPE *m_valuePointer;
	const GAGE_TYPE *m_normalPointer;
	const GAGE_TYPE *m_gradientPointer;
	const GAGE_TYPE *m_gradientMagnitudePointer;
	const GAGE_TYPE *m_hessianPointer;
	const GAGE_TYPE *m_laplacianPointer;
	const GAGE_TYPE *m_hessian1stEigenvaluePointer;
	const GAGE_TYPE *m_hessian2ndEigenvaluePointer;
	const GAGE_TYPE *m_hessian3rdEigenvaluePointer;
	const GAGE_TYPE *m_1stPrincipleCurvaturePointer;
	bool m_isOpen;
	bool m_doClamp;
};

#endif // GAGEADAPTOR_INCLUDED

