/* Copyright 2007 Universidade Federal do Rio Grande do Sul
   Carlos Dietrich - cadietrich@inf.ufrgs.br

   This copy is GPLv2 or later, your choice. */

#include <algorithm>

#include <cmath>
#include <float.h>
#include <iostream>
#include <ErrorManager.h>
#include <MeshQuality.h>

using namespace std;

/**
*/
CMeshQuality::CMeshQuality(void)
{
	Create();
}

/**
*/
CMeshQuality::~CMeshQuality(void)
{
}

/**
*/
bool CMeshQuality::Load(boost::shared_ptr<CVertexPool> mesh)
{
	if (!mesh.get())
	{
		MarkError();

		return false;
	}

	if (mesh->m_vertexArray.empty() || mesh->m_elementArray.empty())
	{
		if (mesh->m_vertexArray.empty()) {
			cout << "vertex array is empty." << endl;
		}
		if (mesh->m_elementArray.empty()) {
			cout << "element array is empty." << endl;
		}
		MarkError();

		return false;
	}

	m_mesh = mesh;

	return true;
}

/**
*/
bool CMeshQuality::SetMeasure(TRIANGLE_QUALITY qualityMeasure)
{
	m_qualityMeasure = qualityMeasure;

	return true;
}

/**
*/
bool CMeshQuality::MeasureQuality(void)
{
	if (!m_mesh.get())
	{
		MarkError();

		return false;
	}

	switch (m_mesh->m_mode)
	{
	case CVertexPool::TRIANGLES:
		MeasureQualityOfTriangleMesh();
		break;
	case CVertexPool::QUADS:
		MeasureQualityOfQuadMesh();
		break;
	default:
		MarkError();

		return false;
	};

	return true;
}

/**
*/
float CMeshQuality::GetVertexQuality(int vertexName) const
{
	if (!m_mesh.get())
	{
		MarkError();

		return 0.0f;
	}

	if (vertexName >= (int)m_qualityArray.size())
	{
		MarkError();

		return 0.0f;
	}

	return m_qualityArray[vertexName];
}

/**
*/
float CMeshQuality::GetMinQuality(void) const
{
	return m_minQuality;
}

/**
*/
float CMeshQuality::GetMaxQuality(void) const
{
	return m_maxQuality;
}

/**
*/
float CMeshQuality::GetAverageQuality(void) const
{
	return m_averageQuality;
}

/**
*/
void CMeshQuality::Create(void)
{
	m_mesh.reset();

	// Adequate default value?
	m_qualityMeasure = RADII_RATIO;

	m_qualityArray.clear();

	m_minQuality = 0.0f;
	m_maxQuality = 1.0f;

	m_averageQuality = 1.0f;

	m_degenerateTriangleCount = 0;
}

/**
*/
bool CMeshQuality::MeasureQualityOfTriangleMesh(void)
{
	int vertexCount,
		triangleCount,
		degenerateTriangleCount,
		triangleElementArray[3],
		i,
		j;
    float *vertexArray,
		*A,
        *B,
        *C,
		quality,
		minQuality,
		maxQuality;
	double qualitySum;
	unsigned int *element;

	if (!m_mesh.get())
	{
		MarkError();

		return false;
	}

	vertexCount = (int)m_mesh->m_vertexArray.size()/m_mesh->m_vertexSize;

	triangleCount = (int)m_mesh->m_elementArray.size()/3;
		
	m_qualityArray.assign(vertexCount, 1.0f);

	minQuality = FLT_MAX;
	maxQuality = FLT_MIN;

	qualitySum = 0.0;

	degenerateTriangleCount = 0;

	element = &m_mesh->m_elementArray[0];

	vertexArray = &m_mesh->m_vertexArray[0];

	for (i=triangleCount; i--; )
    {
		// Triangle quality measures are orientation insensitive.
		for (j=3; j--; )
		{
			triangleElementArray[j] = *element++;

			if (triangleElementArray[j] >= vertexCount)
			{
				MarkError();

				return false;
			}
		}
			
		A = vertexArray + 3*triangleElementArray[0];
		B = vertexArray + 3*triangleElementArray[1];
		C = vertexArray + 3*triangleElementArray[2];

		quality = GetTriangleQuality(A, B, C);

		// This function returns -1 in case or error.
		if (quality >= 0.0f)
		{
			minQuality = min(minQuality, quality);
			maxQuality = max(maxQuality, quality);

			qualitySum += (double)quality;

			// Associates to the vertex the quality of the worst triangle 
			// incident to it.
			for (j=3; j--; )
				m_qualityArray[triangleElementArray[j]] = min(m_qualityArray[triangleElementArray[j]], quality);
		}
		else
			degenerateTriangleCount++;
	}

	m_minQuality = minQuality;
	m_maxQuality = maxQuality;

	m_averageQuality = (float)(qualitySum/(triangleCount - degenerateTriangleCount));

	m_degenerateTriangleCount = degenerateTriangleCount;

	return true;
}

/**
*/
bool CMeshQuality::MeasureQualityOfQuadMesh(void)
{
	int vertexCount,
		quadCount,
		degenerateTriangleCount,
		quadElementArray[4],
		i,
		j,
		k;
    float *vertexArray,
		*A,
        *B,
        *C,
		*D,
		// Treats each quad as two triangles.
		quality[2],
		minQuality,
		maxQuality;
	double qualitySum;
	unsigned int *element;

	if (!m_mesh.get())
	{
		MarkError();

		return false;
	}

	vertexCount = (int)m_mesh->m_vertexArray.size()/m_mesh->m_vertexSize;

	quadCount = (int)m_mesh->m_elementArray.size()/4;
		
	m_qualityArray.assign(vertexCount, 1.0f);

	minQuality = FLT_MAX;
	maxQuality = FLT_MIN;

	qualitySum = 0.0;

	degenerateTriangleCount = 0;

	element = &m_mesh->m_elementArray[0];

	vertexArray = &m_mesh->m_vertexArray[0];

	for (i=quadCount; i--; )
    {
		// Triangle quality measures are orientation insensitive.
		for (j=4; j--; )
		{
			quadElementArray[j] = *element++;

			if (quadElementArray[j] >= vertexCount)
			{
				MarkError();

				return false;
			}
		}
			
		A = vertexArray + 3*quadElementArray[0];
		B = vertexArray + 3*quadElementArray[1];
		C = vertexArray + 3*quadElementArray[2];
		D = vertexArray + 3*quadElementArray[3];

		// CCW orientation.
		quality[0] = GetTriangleQuality(A, B, C);
		quality[1] = GetTriangleQuality(A, C, D);

		for (j=2; j--; )
		{
			// This function returns -1 in case or error.
			if (quality[j] >= 0.0f)
			{
				minQuality = min(minQuality, quality[j]);
				maxQuality = max(maxQuality, quality[j]);

				qualitySum += (double)quality[j];

				// Associates to the vertex the quality of the worst triangle 
				// incident to it.
				for (k=3; k--; )
					m_qualityArray[quadElementArray[k]] = min(m_qualityArray[quadElementArray[k]], quality[j]);
			}
			else
				degenerateTriangleCount++;
		}
	}

	m_minQuality = minQuality;
	m_maxQuality = maxQuality;

	m_averageQuality = (float)(qualitySum/(2*quadCount - degenerateTriangleCount));

	m_degenerateTriangleCount = degenerateTriangleCount;

	return true;
}

/**
*/
float CMeshQuality::GetTriangleQuality(const float *A, const float *B, const float *C) const
{
	float a,
		b,
		c,
		lmin,
		lmax,
		lEpsilon,
		semiperimeter,
		area,
		inradius,
		circumradius,
		quality;
	int i;
		
	a = 0.0f;
	b = 0.0f;
	c = 0.0f;
		
	for (i=3; i--; )
	{
		a += (C[i] - B[i])*(C[i] - B[i]);
		b += (A[i] - C[i])*(A[i] - C[i]);
		c += (B[i] - A[i])*(B[i] - A[i]);
	}
		
    a = sqrtf(a);
    b = sqrtf(b);
	c = sqrtf(c);

#ifndef _MSC_VER
    lmin = std::min(a, std::min(b, c));
	lmax = std::max(a, std::max(b, c));
#else // !_MSC_VER
    lmin = min(a, min(b, c));
	lmax = max(a, max(b, c));
#endif // _MSC_VER

	// Is it better to have an adaptive tolerance?
	lEpsilon = lmax*0.0001f;

	semiperimeter = (a + b + c)*0.5f;

    area = sqrtf(semiperimeter*(semiperimeter - a)*(semiperimeter - b)*(semiperimeter - c));

	// This function returns -1 in case or error.
	if (area < lEpsilon)
		return -1.0f;

    inradius = 2.0f*area/(a + b + c);

	circumradius = (a*b*c)/(4.0f*inradius*semiperimeter);

	switch (m_qualityMeasure) {
		case RADII_RATIO:
			quality = inradius/(0.5f*circumradius);
			break;
		case EDGE_RATIO:
			quality = lmin/lmax;
			break;
		case ASPECT_RATIO:
			quality = inradius/(0.2886751f*lmax);
			break;
		case CIRCUMRADIUS_TO_MAX_EDGE_RATIO:
			quality = lmax/(2.0f*circumradius);
			break;
		case CIRCUMRADIUS_TO_SEMIPERIMETER_RATIO:
			quality = semiperimeter/(2.5980762f*circumradius);
			break;
	}

	return quality;
}

