/* Copyright 2007 Universidade Federal do Rio Grande do Sul
   Carlos Dietrich - cadietrich@inf.ufrgs.br

   This copy is GPLv2 or later, your choice. */

#ifdef _WIN32
#include <windows.h>
#endif // _WIN32

#ifdef __APPLE__
#include <GLUT/glut.h>
#else
#include <GL/gl.h>
#include <GL/glu.h>
#endif


#include <cmath>

#include <PinholeCamera.h>

/**
*/
CPinholeCamera::CPinholeCamera(void)
{
	Create();
}

/**
*/
CPinholeCamera::CPinholeCamera(float fieldOfView, float zNear, float zFar, int windowWidth, int windowHeight)
{
	Create(fieldOfView, zNear, zFar, windowWidth, windowHeight);
}

/**
*/
CPinholeCamera::CPinholeCamera(const CPinholeCamera &pinholeCamera)
{
	Copy(pinholeCamera);
}

/**
*/
CPinholeCamera::~CPinholeCamera(void)
{
}

/**
*/
CPinholeCamera& CPinholeCamera::operator=(const CPinholeCamera &pinholeCamera)
{
	Copy(pinholeCamera);

	return *this;
}

/**
*/
bool CPinholeCamera::Create(float fieldOfView, float zNear, float zFar, int windowWidth, int windowHeight)
{
	// Destroy the stored state?
	//m_type = TRACKBALL;  

	m_fieldOfView = fieldOfView;

	m_zNear = zNear;
	m_zFar = zFar;

	m_windowWidth = windowWidth;
	m_windowHeight = windowHeight;

	m_pitchRotation = 0.0f;
	m_yawRotation = 0.0f;

	m_frontDisplacement = 0.0f;
	m_sideDisplacement = 0.0f;
	m_upDisplacement = 0.0f;

	CalculateProjectionMatrix();

	// Reset the current view matrix?
	//ResetViewMatrix();

	CalculateImagePlaneMatrix();

	return true;
}

/**
*/
void CPinholeCamera::Pitch(float angle)
{
	m_pitchRotation += angle;
}

/**
*/
void CPinholeCamera::Yaw(float angle)
{
	m_yawRotation += angle;
}

/**
*/
void CPinholeCamera::MoveFront(float step)
{
	m_frontDisplacement += step;
}

/**
*/
void CPinholeCamera::MoveSide(float step)
{
	m_sideDisplacement += step;
}

/**
*/
void CPinholeCamera::MoveUp(float step)
{
	m_upDisplacement += step;
}

/**
*/
void CPinholeCamera::ApplyTransform(void)
{
	glMatrixMode(GL_PROJECTION);

	glLoadMatrixf(m_projectionMatrix);

	glMatrixMode(GL_MODELVIEW);

	glLoadMatrixf(m_viewMatrix);

	if (m_type == TRACKBALL)
	{
		glTranslatef(m_sideDisplacement*m_viewMatrix[0] + m_upDisplacement*m_viewMatrix[1] + m_frontDisplacement*m_viewMatrix[2], m_sideDisplacement*m_viewMatrix[4] + m_upDisplacement*m_viewMatrix[5] + m_frontDisplacement*m_viewMatrix[6], m_sideDisplacement*m_viewMatrix[8] + m_upDisplacement*m_viewMatrix[9] + m_frontDisplacement*m_viewMatrix[10]);

		glRotatef(m_pitchRotation, m_viewMatrix[0], m_viewMatrix[4], m_viewMatrix[8]);
		glRotatef(m_yawRotation, m_viewMatrix[1], m_viewMatrix[5], m_viewMatrix[9]);
	}
	else if (m_type == FIRST_PERSON)
	{
		m_centerOfProjection[0] = m_viewMatrix[0]*m_viewMatrix[12] + m_viewMatrix[1]*m_viewMatrix[13] + m_viewMatrix[2]*m_viewMatrix[14];
		m_centerOfProjection[1] = m_viewMatrix[4]*m_viewMatrix[12] + m_viewMatrix[5]*m_viewMatrix[13] + m_viewMatrix[6]*m_viewMatrix[14];
		m_centerOfProjection[2] = m_viewMatrix[8]*m_viewMatrix[12] + m_viewMatrix[9]*m_viewMatrix[13] + m_viewMatrix[10]*m_viewMatrix[14];

		glTranslatef(-m_centerOfProjection[0], -m_centerOfProjection[1], -m_centerOfProjection[2]);

		glRotatef(m_pitchRotation, m_viewMatrix[0], m_viewMatrix[4], m_viewMatrix[8]);
		glRotatef(m_yawRotation, 0.0f, 1.0f, 0.0f);

		m_centerOfProjection[0] += m_sideDisplacement*m_viewMatrix[0] + m_upDisplacement*m_viewMatrix[1] + m_frontDisplacement*m_viewMatrix[2];
		m_centerOfProjection[1] += m_sideDisplacement*m_viewMatrix[4] + m_upDisplacement*m_viewMatrix[5] + m_frontDisplacement*m_viewMatrix[6];
		m_centerOfProjection[2] += m_sideDisplacement*m_viewMatrix[8] + m_upDisplacement*m_viewMatrix[9] + m_frontDisplacement*m_viewMatrix[10];

		glTranslatef(m_centerOfProjection[0], m_centerOfProjection[1], m_centerOfProjection[2]);
	}

	m_pitchRotation = 0.0f;
	m_yawRotation = 0.0f;

	m_frontDisplacement = 0.0f;
	m_sideDisplacement = 0.0f;
	m_upDisplacement = 0.0f;

	glGetFloatv(GL_MODELVIEW_MATRIX, m_viewMatrix);
}

/**
*/
void CPinholeCamera::SetCameraType(TYPE type)
{
	m_type = type;
}

/**
*/
CPinholeCamera::TYPE CPinholeCamera::GetCameraType(void) const
{
	return m_type;
}

/**
*/
void CPinholeCamera::ResetViewMatrix(void)
{
	m_viewMatrix[0] = 1.0f;
	m_viewMatrix[1] = 0.0f;
	m_viewMatrix[2] = 0.0f;
	m_viewMatrix[3] = 0.0f;
	m_viewMatrix[4] = 0.0f;
	m_viewMatrix[5] = 1.0f;
	m_viewMatrix[6] = 0.0f;
	m_viewMatrix[7] = 0.0f;
	m_viewMatrix[8] = 0.0f;
	m_viewMatrix[9] = 0.0f;
	m_viewMatrix[10] = 1.0f;
	m_viewMatrix[11] = 0.0f;
	m_viewMatrix[12] = 0.0f;
	m_viewMatrix[13] = 0.0f;
	m_viewMatrix[14] = 0.0f;
	m_viewMatrix[15] = 1.0f;
}

/**
*/
void CPinholeCamera::SetViewMatrix(const float *viewMatrix)
{
	m_viewMatrix[0] = viewMatrix[0];
	m_viewMatrix[1] = viewMatrix[1];
	m_viewMatrix[2] = viewMatrix[2];
	m_viewMatrix[3] = viewMatrix[3];
	m_viewMatrix[4] = viewMatrix[4];
	m_viewMatrix[5] = viewMatrix[5];
	m_viewMatrix[6] = viewMatrix[6];
	m_viewMatrix[7] = viewMatrix[7];
	m_viewMatrix[8] = viewMatrix[8];
	m_viewMatrix[9] = viewMatrix[9];
	m_viewMatrix[10] = viewMatrix[10];
	m_viewMatrix[11] = viewMatrix[11];
	m_viewMatrix[12] = viewMatrix[12];
	m_viewMatrix[13] = viewMatrix[13];
	m_viewMatrix[14] = viewMatrix[14];
	m_viewMatrix[15] = viewMatrix[15];
}

/**
*/
const float *CPinholeCamera::GetViewMatrix(void) const
{
	return m_viewMatrix;
}

/**
*/
const float *CPinholeCamera::GetProjectionMatrix(void) const
{
	return m_projectionMatrix;
}

/**
*/
const float *CPinholeCamera::GetImagePlaneMatrix(void) const
{
	return m_imagePlaneMatrix;
}

/**
*/
const float *CPinholeCamera::GetCenterOfProjection(void) const
{
	return m_centerOfProjection;
}

/**
*/
void CPinholeCamera::Create(void)
{
	m_type = TRACKBALL;
	
	// Adequate default values?
	m_fieldOfView = 0.0f;

	m_zNear = 0.0f;
	m_zFar = 0.0f;

	m_windowWidth = 0;
	m_windowHeight = 0;
	
	m_pitchRotation = 0.0f;
	m_yawRotation = 0.0f;
	
	m_frontDisplacement = 0.0f;
	m_sideDisplacement = 0.0f;
	m_upDisplacement = 0.0f;

	ResetProjectionMatrix();
	ResetViewMatrix();

	ResetImagePlaneMatrix();
}

/**
*/
void CPinholeCamera::Copy(const CPinholeCamera &pinholeCamera)
{
	m_type = pinholeCamera.m_type;
	
	// Adequate default values?
	m_fieldOfView = pinholeCamera.m_fieldOfView;

	m_zNear = pinholeCamera.m_zNear;
	m_zFar = pinholeCamera.m_zFar;

	m_windowWidth = pinholeCamera.m_windowWidth;
	m_windowHeight = pinholeCamera.m_windowHeight;
	
	m_pitchRotation = pinholeCamera.m_pitchRotation;
	m_yawRotation = pinholeCamera.m_yawRotation;
	
	m_frontDisplacement = pinholeCamera.m_frontDisplacement;
	m_sideDisplacement = pinholeCamera.m_sideDisplacement;
	m_upDisplacement = pinholeCamera.m_upDisplacement;

	m_viewMatrix[0] = pinholeCamera.m_viewMatrix[0];
	m_viewMatrix[1] = pinholeCamera.m_viewMatrix[1];
	m_viewMatrix[2] = pinholeCamera.m_viewMatrix[2];
	m_viewMatrix[3] = pinholeCamera.m_viewMatrix[3];
	m_viewMatrix[4] = pinholeCamera.m_viewMatrix[4];
	m_viewMatrix[5] = pinholeCamera.m_viewMatrix[5];
	m_viewMatrix[6] = pinholeCamera.m_viewMatrix[6];
	m_viewMatrix[7] = pinholeCamera.m_viewMatrix[7];
	m_viewMatrix[8] = pinholeCamera.m_viewMatrix[8];
	m_viewMatrix[9] = pinholeCamera.m_viewMatrix[9];
	m_viewMatrix[10] = pinholeCamera.m_viewMatrix[10];
	m_viewMatrix[11] = pinholeCamera.m_viewMatrix[11];
	m_viewMatrix[12] = pinholeCamera.m_viewMatrix[12];
	m_viewMatrix[13] = pinholeCamera.m_viewMatrix[13];
	m_viewMatrix[14] = pinholeCamera.m_viewMatrix[14];
	m_viewMatrix[15] = pinholeCamera.m_viewMatrix[15];

	m_projectionMatrix[0] = pinholeCamera.m_projectionMatrix[0];
	m_projectionMatrix[1] = pinholeCamera.m_projectionMatrix[1];
	m_projectionMatrix[2] = pinholeCamera.m_projectionMatrix[2];
	m_projectionMatrix[3] = pinholeCamera.m_projectionMatrix[3];
	m_projectionMatrix[4] = pinholeCamera.m_projectionMatrix[4];
	m_projectionMatrix[5] = pinholeCamera.m_projectionMatrix[5];
	m_projectionMatrix[6] = pinholeCamera.m_projectionMatrix[6];
	m_projectionMatrix[7] = pinholeCamera.m_projectionMatrix[7];
	m_projectionMatrix[8] = pinholeCamera.m_projectionMatrix[8];
	m_projectionMatrix[9] = pinholeCamera.m_projectionMatrix[9];
	m_projectionMatrix[10] = pinholeCamera.m_projectionMatrix[10];
	m_projectionMatrix[11] = pinholeCamera.m_projectionMatrix[11];
	m_projectionMatrix[12] = pinholeCamera.m_projectionMatrix[12];
	m_projectionMatrix[13] = pinholeCamera.m_projectionMatrix[13];
	m_projectionMatrix[14] = pinholeCamera.m_projectionMatrix[14];
	m_projectionMatrix[15] = pinholeCamera.m_projectionMatrix[15];

	m_imagePlaneMatrix[0] = pinholeCamera.m_imagePlaneMatrix[0];
	m_imagePlaneMatrix[1] = pinholeCamera.m_imagePlaneMatrix[1];
	m_imagePlaneMatrix[2] = pinholeCamera.m_imagePlaneMatrix[2];
	m_imagePlaneMatrix[3] = pinholeCamera.m_imagePlaneMatrix[3];
	m_imagePlaneMatrix[4] = pinholeCamera.m_imagePlaneMatrix[4];
	m_imagePlaneMatrix[5] = pinholeCamera.m_imagePlaneMatrix[5];
	m_imagePlaneMatrix[6] = pinholeCamera.m_imagePlaneMatrix[6];
	m_imagePlaneMatrix[7] = pinholeCamera.m_imagePlaneMatrix[7];
	m_imagePlaneMatrix[8] = pinholeCamera.m_imagePlaneMatrix[8];

	m_centerOfProjection[0] = pinholeCamera.m_centerOfProjection[0];
	m_centerOfProjection[1] = pinholeCamera.m_centerOfProjection[1];
	m_centerOfProjection[2] = pinholeCamera.m_centerOfProjection[2];
}

/**
*/
void CPinholeCamera::ResetProjectionMatrix(void)
{
	m_projectionMatrix[0] = 1.0f;
	m_projectionMatrix[1] = 0.0f;
	m_projectionMatrix[2] = 0.0f;
	m_projectionMatrix[3] = 0.0f;
	m_projectionMatrix[4] = 0.0f;
	m_projectionMatrix[5] = 1.0f;
	m_projectionMatrix[6] = 0.0f;
	m_projectionMatrix[7] = 0.0f;
	m_projectionMatrix[8] = 0.0f;
	m_projectionMatrix[9] = 0.0f;
	m_projectionMatrix[10] = 1.0f;
	m_projectionMatrix[11] = 0.0f;
	m_projectionMatrix[12] = 0.0f;
	m_projectionMatrix[13] = 0.0f;
	m_projectionMatrix[14] = 0.0f;
	m_projectionMatrix[15] = 1.0f;
}

/**
*/
void CPinholeCamera::CalculateProjectionMatrix(void)
{
	glMatrixMode(GL_PROJECTION);

	glPushMatrix();

	glLoadIdentity();

	gluPerspective(m_fieldOfView, (GLdouble)m_windowWidth/m_windowHeight, m_zNear, m_zFar);

	glGetFloatv(GL_PROJECTION_MATRIX, m_projectionMatrix);

	glPopMatrix();

	glMatrixMode(GL_MODELVIEW);
}

/**
*/
void CPinholeCamera::ResetImagePlaneMatrix(void)
{
	m_imagePlaneMatrix[0] = 1.0f;
	m_imagePlaneMatrix[1] = 0.0f;
	m_imagePlaneMatrix[2] = 0.0f;
	m_imagePlaneMatrix[3] = 0.0f;
	m_imagePlaneMatrix[4] = 1.0f;
	m_imagePlaneMatrix[5] = 0.0f;
	m_imagePlaneMatrix[6] = 0.0f;
	m_imagePlaneMatrix[7] = 0.0f;
	m_imagePlaneMatrix[8] = 1.0f;
}

/**
*/
void CPinholeCamera::CalculateImagePlaneMatrix(void)
{
	float w,
		h,
		as,
		bs;
	
	if (m_windowWidth > m_windowHeight)
	{
		w = 2.0f*m_zNear*(float)tan(m_fieldOfView*0.00872664625);
		h = ((float)m_windowHeight/m_windowWidth)*w;
	}
	else
	{
		h = 2.0f*m_zNear*(float)tan(m_fieldOfView*0.00872664625);
		w = ((float)m_windowWidth/m_windowHeight)*h;
	}
	
	as = w/m_windowWidth;
	
	m_imagePlaneMatrix[0] = m_viewMatrix[0]*as;
	m_imagePlaneMatrix[1] = m_viewMatrix[4]*as;
	m_imagePlaneMatrix[2] = m_viewMatrix[8]*as;
	
	bs = h/m_windowHeight;
	
	m_imagePlaneMatrix[3] = -m_viewMatrix[1]*bs;
	m_imagePlaneMatrix[4] = -m_viewMatrix[5]*bs;
	m_imagePlaneMatrix[5] = -m_viewMatrix[9]*bs;
	
	m_imagePlaneMatrix[6] = m_viewMatrix[2]*m_zNear - 0.5f*(m_imagePlaneMatrix[0]*m_windowWidth + m_imagePlaneMatrix[3]*m_windowHeight);
	m_imagePlaneMatrix[7] = m_viewMatrix[6]*m_zNear - 0.5f*(m_imagePlaneMatrix[1]*m_windowWidth + m_imagePlaneMatrix[4]*m_windowHeight);
	m_imagePlaneMatrix[8] = m_viewMatrix[10]*m_zNear - 0.5f*(m_imagePlaneMatrix[2]*m_windowWidth + m_imagePlaneMatrix[5]*m_windowHeight);
}

